@@ -17,7 +17,6 @@ import importlib.util
 import io
 import logging
 import os
-import platform
 import shutil
 import subprocess
 import sys
@@ -28,7 +27,6 @@ from pathlib import Path
 from unittest.mock import patch
 
 import logassert
-import yaml
 
 from ops.charm import (
     ActionEvent,
@@ -46,32 +44,32 @@ from ops.charm import (
     RelationDepartedEvent,
     RelationEvent,
     RelationJoinedEvent,
+    SecretChangedEvent,
+    SecretEvent,
+    SecretExpiredEvent,
+    SecretRemoveEvent,
+    SecretRotateEvent,
     StartEvent,
     StorageAttachedEvent,
     UpdateStatusEvent,
     UpgradeCharmEvent,
     WorkloadEvent,
 )
-from ops.framework import Framework, StoredStateData
+from ops.framework import Framework, LifecycleEvent, StoredStateData
 from ops.main import CHARM_STATE_FILE, _should_use_controller_storage, main
 from ops.storage import SQLiteStorage
 from ops.version import version
 
+from .charms.test_main.src.charm import MyCharmEvents
 from .test_helpers import fake_script, fake_script_calls
 
-is_windows = platform.system() == 'Windows'
-
 # This relies on the expected repository structure to find a path to
 # source of the charm under test.
-TEST_CHARM_DIR = Path(__file__ + '/../charms/test_main').resolve()
+TEST_CHARM_DIR = Path(f"{__file__}/../charms/test_main").resolve()
 
 VERSION_LOGLINE = [
     'juju-log', '--log-level', 'DEBUG', '--',
-    'Operator Framework {} up and running.'.format(version),
-]
-SLOW_YAML_LOGLINE = [
-    'juju-log', '--log-level', 'DEBUG', '--',
-    'yaml does not have libyaml extensions, using slower pure Python yaml loader',
+    f'Operator Framework {version} up and running.',
 ]
 
 logger = logging.getLogger(__name__)
@@ -85,7 +83,8 @@ class EventSpec:
     def __init__(self, event_type, event_name, env_var=None,
                  relation_id=None, remote_app=None, remote_unit=None,
                  model_name=None, set_in_env=None, workload_name=None,
-                 departing_unit_name=None):
+                 departing_unit_name=None, secret_id=None, secret_label=None,
+                 secret_revision=None):
         self.event_type = event_type
         self.event_name = event_name
         self.env_var = env_var
@@ -96,12 +95,14 @@ class EventSpec:
         self.model_name = model_name
         self.set_in_env = set_in_env
         self.workload_name = workload_name
+        self.secret_id = secret_id
+        self.secret_label = secret_label
+        self.secret_revision = secret_revision
 
 
 @patch('ops.main.setup_root_logging', new=lambda *a, **kw: None)
 class CharmInitTestCase(unittest.TestCase):
 
-    @unittest.skipIf(sys.version_info < (3, 7), "no breakpoint builtin for Python < 3.7")
     @patch('sys.stderr', new_callable=io.StringIO)
     def test_breakpoint(self, fake_stderr):
         class MyCharm(CharmBase):
@@ -109,19 +110,18 @@ class CharmInitTestCase(unittest.TestCase):
         self._check(MyCharm, extra_environ={'JUJU_DEBUG_AT': 'all'})
 
         with patch('pdb.Pdb.set_trace') as mock:
-            breakpoint()        # noqa: F821 ('undefined name' in <3.7)
+            breakpoint()
 
         self.assertEqual(mock.call_count, 1)
         self.assertIn('Starting pdb to debug charm operator', fake_stderr.getvalue())
 
-    @unittest.skipIf(sys.version_info < (3, 7), "no breakpoint builtin for Python < 3.7")
     def test_no_debug_breakpoint(self):
         class MyCharm(CharmBase):
             pass
         self._check(MyCharm, extra_environ={'JUJU_DEBUG_AT': ''})
 
         with patch('pdb.Pdb.set_trace') as mock:
-            breakpoint()        # noqa: F821 ('undefined name' in <3.7)
+            breakpoint()
 
         self.assertEqual(mock.call_count, 0)
 
@@ -144,10 +144,7 @@ class CharmInitTestCase(unittest.TestCase):
                             fh.write(b'name: test')
                         mock_charmdir.return_value = tmpdirname
 
-                        with warnings.catch_warnings(record=True) as warnings_cm:
-                            main(charm_class, **kwargs)
-
-        return warnings_cm
+                        main(charm_class, **kwargs)
 
     def test_init_signature_passthrough(self):
         class MyCharm(CharmBase):
@@ -155,22 +152,19 @@ class CharmInitTestCase(unittest.TestCase):
             def __init__(self, *args):
                 super().__init__(*args)
 
-        warn_cm = self._check(MyCharm)
-        self.assertFalse(warn_cm)
+        with warnings.catch_warnings(record=True) as warn_cm:
+            self._check(MyCharm)
+        self.assertEqual(warn_cm, [])
 
-    def test_init_signature_both_arguments(self):
+    def test_init_signature_old_key_argument(self):
         class MyCharm(CharmBase):
 
             def __init__(self, framework, somekey):
                 super().__init__(framework, somekey)
 
-        warn_cm = self._check(MyCharm)
-        self.assertEqual(len(warn_cm), 1)
-        (warn,) = warn_cm
-        self.assertTrue(issubclass(warn.category, DeprecationWarning))
-        self.assertEqual(str(warn.message), (
-            "the second argument, 'key', has been deprecated and will be removed "
-            "after the 0.7 release"))
+        # Support for "key" has been deprecated since ops 0.7 and was removed in 2.0
+        with self.assertRaises(TypeError):
+            self._check(MyCharm)
 
     def test_init_signature_only_framework(self):
         class MyCharm(CharmBase):
@@ -178,8 +172,9 @@ class CharmInitTestCase(unittest.TestCase):
             def __init__(self, framework):
                 super().__init__(framework)
 
-        warn_cm = self._check(MyCharm)
-        self.assertFalse(warn_cm)
+        with warnings.catch_warnings(record=True) as warn_cm:
+            self._check(MyCharm)
+        self.assertEqual(warn_cm, [])
 
     def test_storage_no_storage(self):
         # here we patch juju_backend_available so it refuses to set it up
@@ -197,6 +192,13 @@ class CharmInitTestCase(unittest.TestCase):
             with self.assertRaisesRegex(FileNotFoundError, 'state-get'):
                 self._check(CharmBase, use_juju_for_storage=True)
 
+    def test_controller_storage_deprecated(self):
+        with patch('ops.storage.juju_backend_available') as juju_backend_available:
+            juju_backend_available.return_value = True
+            with self.assertWarnsRegex(DeprecationWarning, 'Controller storage'):
+                with self.assertRaisesRegex(FileNotFoundError, 'state-get'):
+                    self._check(CharmBase, use_juju_for_storage=True)
+
 
 @patch('sys.argv', new=("hooks/config-changed",))
 @patch('ops.main.setup_root_logging', new=lambda *a, **kw: None)
@@ -328,7 +330,7 @@ class _TestMain(abc.ABC):
         for action_name in ('start', 'foo-bar', 'get-model-name', 'get-status'):
             self._setup_entry_point(actions_dir, action_name)
 
-    def _read_and_clear_state(self):
+    def _read_and_clear_state(self, event_name: str):
         if self.CHARM_STATE_FILE.stat().st_size:
             storage = SQLiteStorage(self.CHARM_STATE_FILE)
             with (self.JUJU_CHARM_DIR / 'metadata.yaml').open() as m:
@@ -338,9 +340,9 @@ class _TestMain(abc.ABC):
                         meta = CharmMeta.from_yaml(m, a)
                 else:
                     meta = CharmMeta.from_yaml(m)
-            framework = Framework(storage, self.JUJU_CHARM_DIR, meta, None)
+            framework = Framework(storage, self.JUJU_CHARM_DIR, meta, None, event_name)
 
-            class ThisCharmEvents(CharmEvents):
+            class ThisCharmEvents(MyCharmEvents):
                 pass
 
             class Charm(self.charm_module.Charm):
@@ -370,6 +372,15 @@ class _TestMain(abc.ABC):
         })
         if event_spec.set_in_env is not None:
             env.update(event_spec.set_in_env)
+        if issubclass(event_spec.event_type, SecretEvent):
+            env.update({
+                'JUJU_SECRET_ID': event_spec.secret_id,
+                'JUJU_SECRET_LABEL': event_spec.secret_label or '',
+            })
+        if issubclass(event_spec.event_type, (SecretRemoveEvent, SecretExpiredEvent)):
+            env.update({
+                'JUJU_SECRET_REVISION': str(event_spec.secret_revision or ''),
+            })
         if issubclass(event_spec.event_type, RelationEvent):
             rel_name = event_spec.event_name.split('_')[0]
             env.update({
@@ -415,7 +426,7 @@ class _TestMain(abc.ABC):
             env['JUJU_MODEL_NAME'] = event_spec.model_name
 
         self._call_event(Path(event_dir, event_filename), env)
-        return self._read_and_clear_state()
+        return self._read_and_clear_state(event_spec.event_name)
 
     def test_event_reemitted(self):
         # First run "install" to make sure all hooks are set up.
@@ -547,6 +558,37 @@ class _TestMain(abc.ABC):
             EventSpec(PebbleReadyEvent, 'test_pebble_ready',
                       workload_name='test'),
             {'container_name': 'test'},
+        ), (
+            EventSpec(SecretChangedEvent, 'secret_changed',
+                      secret_id='secret:12345',
+                      secret_label='foo'),
+            {'id': 'secret:12345',
+             'label': 'foo',
+             'revision': 42}
+        ), (
+            EventSpec(SecretRotateEvent, 'secret_rotate',
+                      secret_id='secret:12345',
+                      secret_label='foo',
+                      secret_revision='42'),
+            {'id': 'secret:12345',
+             'label': 'foo',
+             'revision': 42}
+        ), (
+            EventSpec(SecretRemoveEvent, 'secret_remove',
+                      secret_id='secret:12345',
+                      secret_label='foo',
+                      secret_revision='42'),
+            {'id': 'secret:12345',
+             'label': 'foo',
+             'revision': 42}
+        ), (
+            EventSpec(SecretExpiredEvent, 'secret_expired',
+                      secret_id='secret:12345',
+                      secret_label='foo',
+                      secret_revision='42'),
+            {'id': 'secret:12345',
+             'label': 'foo',
+             'revision': 42}
         )]
 
         logger.debug('Expected events %s', events_under_test)
@@ -558,7 +600,7 @@ class _TestMain(abc.ABC):
         for event_spec, expected_event_data in events_under_test:
             state = self._simulate_event(event_spec)
 
-            state_key = 'on_' + event_spec.event_name
+            state_key = f"on_{event_spec.event_name}"
             handled_events = getattr(state, state_key, [])
 
             # Make sure that a handler for that event was called once.
@@ -570,7 +612,7 @@ class _TestMain(abc.ABC):
             self.assertEqual(list(state.observed_event_types), [event_spec.event_type.__name__])
 
             if event_spec.event_name in expected_event_data:
-                self.assertEqual(state[event_spec.event_name + '_data'],
+                self.assertEqual(state[f"{event_spec.event_name}_data"],
                                  expected_event_data[event_spec.event_name])
 
     def test_event_not_implemented(self):
@@ -604,22 +646,29 @@ class _TestMain(abc.ABC):
 
         expected = [
             VERSION_LOGLINE,
-            ['juju-log', '--log-level', 'DEBUG', '--',
-             'Using local storage: {} already exists'.format(self.CHARM_STATE_FILE)],
             ['juju-log', '--log-level', 'DEBUG', '--', 'Emitting Juju event collect_metrics.'],
             ['add-metric', '--labels', 'bar=4.2', 'foo=42'],
         ]
-        if not yaml.__with_libyaml__:
-            expected.insert(1, SLOW_YAML_LOGLINE)
         calls = fake_script_calls(self)
 
-        if self.has_dispatch:
-            expected.insert(1, [
-                'juju-log', '--log-level', 'DEBUG', '--',
-                'Legacy {} does not exist.'.format(Path('hooks/collect-metrics'))])
-
         self.assertEqual(calls, expected)
 
+    def test_custom_event(self):
+        self._simulate_event(EventSpec(InstallEvent, 'install'))
+        # Clear the calls during 'install'
+        fake_script_calls(self, clear=True)
+        self._simulate_event(EventSpec(UpdateStatusEvent, 'update-status',
+                                       set_in_env={'EMIT_CUSTOM_EVENT': "1"}))
+
+        expected = [
+            VERSION_LOGLINE,
+            ['juju-log', '--log-level', 'DEBUG', '--', 'Emitting Juju event update_status.'],
+            ['juju-log', '--log-level', 'DEBUG', '--', 'Emitting custom event '
+                                                       '<CustomEvent via Charm/on/custom[5]>.'],
+        ]
+        calls = fake_script_calls(self)
+        self.assertEqual(expected, calls)
+
     def test_logger(self):
         fake_script(self, 'action-get', "echo '{}'")
 
@@ -647,7 +696,6 @@ class _TestMain(abc.ABC):
             self._simulate_event(event_spec)
             self.assertIn(calls, fake_script_calls(self, clear=True))
 
-    @unittest.skipIf(is_windows, 'TODO windows multiline args are hard')
     def test_excepthook(self):
         with self.assertRaises(subprocess.CalledProcessError):
             self._simulate_event(EventSpec(InstallEvent, 'install',
@@ -656,17 +704,8 @@ class _TestMain(abc.ABC):
         calls = [' '.join(i) for i in fake_script_calls(self)]
 
         self.assertEqual(calls.pop(0), ' '.join(VERSION_LOGLINE))
-
-        if self.has_dispatch:
-            self.assertEqual(
-                calls.pop(0),
-                'juju-log --log-level DEBUG -- Legacy {} does not exist.'.format(
-                    Path("hooks/install")))
-
-        if not yaml.__with_libyaml__:
-            self.assertEqual(calls.pop(0), ' '.join(SLOW_YAML_LOGLINE))
-
-        self.assertRegex(calls.pop(0), 'Using local storage: not a kubernetes charm')
+        self.assertRegex(calls.pop(0), 'Using local storage: not a Kubernetes podspec charm')
+        self.assertRegex(calls.pop(0), 'Initializing SQLite local storage: ')
 
         self.maxDiff = None
         self.assertRegex(
@@ -677,7 +716,7 @@ class _TestMain(abc.ABC):
             '    raise RuntimeError."failing as requested".\n'
             'RuntimeError: failing as requested'
         )
-        self.assertEqual(len(calls), 1, "expected 1 call, but got extra: {}".format(calls[1:]))
+        self.assertEqual(len(calls), 1, f"expected 1 call, but got extra: {calls[1:]}")
 
     def test_sets_model_name(self):
         self._prepare_actions()
@@ -770,26 +809,26 @@ class TestMainWithNoDispatch(_TestMain, unittest.TestCase):
 
     def test_setup_event_links(self):
         """Test auto-creation of symlinks caused by initial events."""
-        all_event_hooks = ['hooks/' + e.replace("_", "-")
-                           for e in self.charm_module.Charm.on.events().keys()]
+        all_event_hooks = [f"hooks/{name.replace('_', '-')}"
+                           for name, event_source in self.charm_module.Charm.on.events().items()
+                           if issubclass(event_source.event_type, LifecycleEvent)]
+
         initial_events = {
             EventSpec(InstallEvent, 'install'),
             EventSpec(StorageAttachedEvent, 'disks-storage-attached'),
             EventSpec(StartEvent, 'start'),
             EventSpec(UpgradeCharmEvent, 'upgrade-charm'),
         }
-        initial_hooks = {'hooks/' + ev.event_name for ev in initial_events}
+        initial_hooks = {f"hooks/{ev.event_name}" for ev in initial_events}
 
         def _assess_event_links(event_spec):
             self.assertTrue(self.hooks_dir / event_spec.event_name in self.hooks_dir.iterdir())
             for event_hook in all_event_hooks:
                 hook_path = self.JUJU_CHARM_DIR / event_hook
-                self.assertTrue(hook_path.exists(), 'Missing hook: ' + event_hook)
+                self.assertTrue(hook_path.exists(), f"Missing hook: {event_hook}")
                 if self.hooks_are_symlinks:
                     self.assertTrue(hook_path.is_symlink())
-                    if not is_windows:
-                        # TODO(benhoyt): fix this now that tests are running on GitHub Actions
-                        self.assertEqual(os.readlink(str(hook_path)), self.charm_exec_path)
+                    self.assertEqual(os.readlink(str(hook_path)), self.charm_exec_path)
                 elif event_hook in initial_hooks:
                     self.assertFalse(hook_path.is_symlink())
                 else:
@@ -844,7 +883,7 @@ class _TestMainWithDispatch(_TestMain):
 
         Symlink creation caused by initial events should _not_ happen when using dispatch.
         """
-        all_event_hooks = ['hooks/' + e.replace("_", "-")
+        all_event_hooks = [f"hooks/{e.replace('_', '-')}"
                            for e in self.charm_module.Charm.on.events().keys()]
         initial_events = {
             EventSpec(InstallEvent, 'install'),
@@ -857,7 +896,7 @@ class _TestMainWithDispatch(_TestMain):
             self.assertNotIn(self.hooks_dir / event_spec.event_name, self.hooks_dir.iterdir())
             for event_hook in all_event_hooks:
                 self.assertFalse((self.JUJU_CHARM_DIR / event_hook).exists(),
-                                 'Spurious hook: ' + event_hook)
+                                 f"Spurious hook: {event_hook}")
 
         for initial_event in initial_events:
             self._setup_charm_dir()
@@ -880,19 +919,18 @@ class _TestMainWithDispatch(_TestMain):
         expected = [
             VERSION_LOGLINE,
             ['juju-log', '--log-level', 'INFO', '--',
-             'Running legacy {}.'.format(hook)],
+             f'Running legacy {hook}.'],
             ['juju-log', '--log-level', 'DEBUG', '--',
-             'Legacy {} exited with status 0.'.format(hook)],
+             f'Legacy {hook} exited with status 0.'],
             ['juju-log', '--log-level', 'DEBUG', '--',
-             'Using local storage: not a kubernetes charm'],
+             'Using local storage: not a Kubernetes podspec charm'],
             ['juju-log', '--log-level', 'DEBUG', '--',
              'Emitting Juju event install.'],
         ]
-        if not yaml.__with_libyaml__:
-            expected.insert(3, SLOW_YAML_LOGLINE)
-        self.assertEqual(fake_script_calls(self), expected)
+        calls = fake_script_calls(self)
+        self.assertRegex(' '.join(calls.pop(-2)), 'Initializing SQLite local storage: ')
+        self.assertEqual(calls, expected)
 
-    @unittest.skipIf(is_windows, "this is UNIXish; TODO: write equivalent windows test")
     def test_non_executable_hook_and_dispatch(self):
         (self.hooks_dir / "install").write_text("")
         state = self._simulate_event(EventSpec(InstallEvent, 'install'))
@@ -904,13 +942,13 @@ class _TestMainWithDispatch(_TestMain):
             ['juju-log', '--log-level', 'WARNING', '--',
              'Legacy hooks/install exists but is not executable.'],
             ['juju-log', '--log-level', 'DEBUG', '--',
-             'Using local storage: not a kubernetes charm'],
+             'Using local storage: not a Kubernetes podspec charm'],
             ['juju-log', '--log-level', 'DEBUG', '--',
              'Emitting Juju event install.'],
         ]
-        if not yaml.__with_libyaml__:
-            expected.insert(2, SLOW_YAML_LOGLINE)
-        self.assertEqual(fake_script_calls(self), expected)
+        calls = fake_script_calls(self)
+        self.assertRegex(' '.join(calls.pop(-2)), 'Initializing SQLite local storage: ')
+        self.assertEqual(calls, expected)
 
     def test_hook_and_dispatch_with_failing_hook(self):
         self.stdout = self.stderr = tempfile.TemporaryFile()
@@ -930,17 +968,15 @@ class _TestMainWithDispatch(_TestMain):
         hook = Path('hooks/install')
         expected = [
             VERSION_LOGLINE,
-            ['juju-log', '--log-level', 'INFO', '--', 'Running legacy {}.'.format(hook)],
+            ['juju-log', '--log-level', 'INFO', '--', f'Running legacy {hook}.'],
             ['juju-log', '--log-level', 'WARNING', '--',
-             'Legacy {} exited with status 42.'.format(hook)],
+             f'Legacy {hook} exited with status 42.'],
         ]
         self.assertEqual(calls, expected)
 
     def test_hook_and_dispatch_but_hook_is_dispatch(self):
         event = EventSpec(InstallEvent, 'install')
         hook_path = self.hooks_dir / 'install'
-        if is_windows:
-            hook_path = hook_path.with_suffix('.bat')
         for ((rel, ind), path) in {
                 # relative and indirect
                 (True, True): Path('../dispatch'),
@@ -952,8 +988,6 @@ class _TestMainWithDispatch(_TestMain):
                 (False, True): self.JUJU_CHARM_DIR / 'dispatch',
         }.items():
             with self.subTest(path=path, rel=rel, ind=ind):
-                if is_windows:
-                    path = path.with_suffix('.sh')
                 # sanity check
                 self.assertEqual(path.is_absolute(), not rel)
                 self.assertEqual(path.with_suffix('').name == 'dispatch', ind)
@@ -968,7 +1002,6 @@ class _TestMainWithDispatch(_TestMain):
                 finally:
                     hook_path.unlink()
 
-    @unittest.skipIf(is_windows, "this needs rethinking on Windows")
     def test_hook_and_dispatch_but_hook_is_dispatch_copy(self):
         hook_path = self.hooks_dir / 'install'
         path = (self.hooks_dir / self.charm_exec_path).resolve()
@@ -984,26 +1017,23 @@ class _TestMainWithDispatch(_TestMain):
         expected = [
             VERSION_LOGLINE,
             ['juju-log', '--log-level', 'INFO', '--',
-             'Running legacy {}.'.format(hook)],
+             f'Running legacy {hook}.'],
             VERSION_LOGLINE,    # because it called itself
             ['juju-log', '--log-level', 'DEBUG', '--',
-             'Charm called itself via {}.'.format(hook)],
+             f'Charm called itself via {hook}.'],
             ['juju-log', '--log-level', 'DEBUG', '--',
-             'Legacy {} exited with status 0.'.format(hook)],
+             f'Legacy {hook} exited with status 0.'],
             ['juju-log', '--log-level', 'DEBUG', '--',
-             'Using local storage: not a kubernetes charm'],
+             'Using local storage: not a Kubernetes podspec charm'],
             ['juju-log', '--log-level', 'DEBUG', '--',
              'Emitting Juju event install.'],
         ]
-        if not yaml.__with_libyaml__:
-            expected.insert(5, SLOW_YAML_LOGLINE)
-        self.assertEqual(fake_script_calls(self), expected)
+        calls = fake_script_calls(self)
+        self.assertRegex(' '.join(calls.pop(-2)), 'Initializing SQLite local storage: ')
+
+        self.assertEqual(calls, expected)
 
 
-# NOTE
-#  AIUI On windows dispatch must be a script (see TestMainWithDispatchAsScript),
-#  because Juju won't call python even if we rename dispatch to dispatch.py
-@unittest.skipIf(is_windows, "Juju on windows won't make this work (see note)")
 class TestMainWithDispatch(_TestMainWithDispatch, unittest.TestCase):
     def _setup_entry_point(self, directory, entry_point):
         path = self.JUJU_CHARM_DIR / 'dispatch'
@@ -1066,17 +1096,10 @@ class TestMainWithDispatchAsScript(_TestMainWithDispatch, unittest.TestCase):
 
     has_dispatch = True
 
-    if is_windows:
-        suffix = '.BAT'
-        script = '@ECHO OFF\n"{}" "{}"\n'
-    else:
-        suffix = ''
-        script = '#!/bin/sh\nexec "{}" "{}"\n'
-
     def _setup_entry_point(self, directory, entry_point):
-        path = (self.JUJU_CHARM_DIR / 'dispatch').with_suffix(self.suffix)
+        path = (self.JUJU_CHARM_DIR / 'dispatch')
         if not path.exists():
-            path.write_text(self.script.format(
+            path.write_text('#!/bin/sh\nexec "{}" "{}"\n'.format(
                 sys.executable,
                 self.JUJU_CHARM_DIR / 'src/charm.py'))
             path.chmod(0o755)
@@ -1124,7 +1147,7 @@ class TestMainWithDispatchAsScript(_TestMainWithDispatch, unittest.TestCase):
             echo '["disks/0"]'
             """,
         )
-        dispatch = (self.JUJU_CHARM_DIR / 'dispatch').with_suffix(self.suffix)
+        dispatch = (self.JUJU_CHARM_DIR / 'dispatch')
         subprocess.check_call([str(dispatch)], env=env, cwd=str(self.JUJU_CHARM_DIR))
 
 
@@ -1148,10 +1171,9 @@ class TestStorageHeuristics(unittest.TestCase):
         meta = CharmMeta.from_yaml("series: [ecs]")
         with patch.dict(os.environ, {"JUJU_VERSION": "2.8"}):
             self.assertFalse(_should_use_controller_storage(Path("/xyzzy"), meta))
-            self.assertLogged('Using local storage: not a kubernetes charm')
+            self.assertLogged('Using local storage: not a Kubernetes podspec charm')
 
     def test_not_if_already_local(self):
         meta = CharmMeta.from_yaml("series: [kubernetes]")
         with patch.dict(os.environ, {"JUJU_VERSION": "2.8"}), tempfile.NamedTemporaryFile() as fd:
             self.assertFalse(_should_use_controller_storage(Path(fd.name), meta))
-            self.assertLogged('Using local storage: {} already exists'.format(fd.name))
